// Source Code for KNL Programming Tutorial
// KISTI Intel(R) Parallel Programming Center 
#include <stdio.h>
#include <string.h>
#include <omp.h>
//#define SIZE 4096
#define SIZE 2048
#ifdef HAVE_CBLAS
#include "mkl.h"
#endif 
int main(int argc, char *argv[]) {
	double time;
	double *A = (double*)_mm_malloc(sizeof(double)*SIZE*SIZE, 64);
	double *B = (double*)_mm_malloc(sizeof(double)*SIZE*SIZE, 64);
	double *C = (double*)_mm_malloc(sizeof(double)*SIZE*SIZE, 64);
	#pragma omp parallel
	{
		#pragma omp master
		{
			printf("\tomp_num_thread: %d \n", omp_get_num_threads());
		}
	}
	#pragma omp parallel for
	for(int i=0; i<SIZE; i++) {
		#pragma omp simd
		#pragma vector aligned
		//#pragma unroll(8)
		for(int j=0; j<SIZE; j++) {
			A[i*SIZE+j] = (double)(i + j);
			B[i*SIZE+j] = (double)(j - i);
		}
	}
/////////////////////////////////////////////////
//	memset(C, 0, sizeof(double)*SIZE*SIZE);
//
//	time = -omp_get_wtime();
//	#pragma omp parallel for 
//	for(int i=0; i<SIZE; i++) {
//		#pragma omp simd
//		#pragma vector aligned
//		for(int j=0; j<SIZE; j++) {
//			double sum = 0;
//			for(int k=0; k<SIZE; k++) {
//				sum += A[i*SIZE+k] * B[k*SIZE+j];
//			}
//			C[i*SIZE+j] = sum;
//		}
//	}
//	time += omp_get_wtime();
//	printf("\ti-j-k MMmul time: %lf (secs)\n", time);
//	printf("\t\tlast element: %lf\n\n", C[(SIZE-1)*SIZE+SIZE-1]);
/////////////////////////////////////////////////
	memset(C, 0, sizeof(double)*SIZE*SIZE);
#ifndef HAVE_CBLAS
	time = -omp_get_wtime();
	//#pragma omp parallel for collapse(2)
	for(int i=0; i<SIZE; i++) {
		for(int k=0; k<SIZE; k++) {
			double A_val = A[i*SIZE+k];
			//#pragma omp simd
			//#pragma vector aligned
			//#pragma nounroll
			//#pragma unroll(8)
			for(int j=0; j<SIZE; j++) {
				C[i*SIZE+j] += A_val * B[k*SIZE+j];
			}
		}
	}
	time += omp_get_wtime();
	printf("\tMMmul time: %lf (secs)\n", time);
	//printf("\ti-k-j MMmul time: %lf (secs)\n", time);
	printf("\t\tlast element: %lf\n\n", C[(SIZE-1)*SIZE+SIZE-1]);
/////////////////////////////////////////////////
#else
	memset(C, 0, sizeof(double)*SIZE*SIZE);

	time = -omp_get_wtime();
	cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,SIZE, SIZE, SIZE, 1.0, A, SIZE, B, SIZE, 0.0, C, SIZE);
	time += omp_get_wtime();
	printf("\tblas time: %lf (secs)\n", time);
	printf("\t\tlast element: %lf\n\n", C[(SIZE-1)*SIZE+SIZE-1]);
#endif
	_mm_free(A);
	_mm_free(B);
	_mm_free(C);

	return 0;
}
