// Source Code for KNL Programming Tutorial
// KISTI Intel(R) Parallel Programming Center 
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#define N 160000000
#define Nnz 64000

int main(int argc, char* argv[])
{
        #pragma omp parallel
        {
                #pragma omp master
                {
                        printf("\tomp_num_thread: %d \n", omp_get_num_threads());
                }
        }

	int* index = (int*)malloc(sizeof(int)*Nnz);
	double* sVec = (double*)malloc(sizeof(double)*Nnz);
	double* dVec = (double*)malloc(sizeof(double)*N);
	double* rVec = (double*)malloc(sizeof(double)*Nnz);
	double* temp = (double*)malloc(sizeof(double)*Nnz);
	double sum, time;

	#pragma omp parallel for
	for(int i=0; i<N; i++)
		dVec[i] = (double)i;
	#pragma omp parallel for
	for(int i=0; i<Nnz; i++)
	{
		sVec[i] = (double)i;
		index[i] = i*(int)(N/Nnz);
		rVec[i] = 0.0;
	}
/////////////////////////////////////////////////
	time = -omp_get_wtime();
	for(int ii=0; ii<10000; ii++)
	{
		#pragma omp simd
		for(int i=0; i<Nnz; i++)
			rVec[i] = dVec[index[i]] * sVec[i];
	}
	time += omp_get_wtime();
	printf("\t1 VVdot time: %lf (secs)\n", time);
	sum = 0.0;
	#pragma omp simd reduction(+:sum)
	for(int i=0; i<Nnz; i++)
		sum += rVec[i];
	printf("\t\tsum: %lf\n\n", sum);
/////////////////////////////////////////////////
	#pragma omp parallel for
	for(int i=0; i<Nnz; i++)
		temp[i] = dVec[index[i]];
	time = -omp_get_wtime();
	for(int ii=0; ii<10000; ii++)
	{
		#pragma omp simd
		for(int i=0; i<Nnz; i++)
			rVec[i] = temp[i] * sVec[i];
	}
	time += omp_get_wtime();
	printf("\t2 VVdot time: %lf (secs)\n", time);
	sum = 0.0;
	#pragma omp simd reduction(+:sum)
	for(int i=0; i<Nnz; i++)
		sum += rVec[i];
	printf("\t\tsum: %lf\n", sum);
/////////////////////////////////////////////////
	free(index);
	free(sVec);
	free(dVec);
	free(rVec);
	free(temp);
}
