// Source Code for KNL Programming Tutorial
// KISTI Intel(R) Parallel Programming Center 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#define NTRIAL 100000000

int main()
{
    double *x = malloc(sizeof *x * NTRIAL);
    double *y = malloc(sizeof *y * NTRIAL);
    int *z = malloc(sizeof *z * NTRIAL);
    double time;
    unsigned short seed[3] = {155, 0, 155};
    seed48(&seed[0]);

    int count = 0;
    time = -omp_get_wtime();
    for (size_t i = 0; i < NTRIAL; i++) {
        double xx = 2. * drand48() - 1.; 
        double yy = 2. * drand48() - 1.; 
        if (xx*xx + yy*yy <= 1.) count++;
    }
    time += omp_get_wtime();
    printf("value: %e\ncore-time: %f(secs)\n\n", 4.*count/NTRIAL, time);

    time = -omp_get_wtime();
    count = 0;
    for (size_t i = 0; i < NTRIAL; i++)
        x[i] = 2. * drand48() - 1.;
    for (size_t i = 0; i < NTRIAL; i++)
        y[i] = 2. * drand48() - 1.;
    for (size_t i = 0; i < NTRIAL; i++)
        if (x[i]*x[i] + y[i]*y[i] <= 1.) count++;
    time += omp_get_wtime();
    printf("value: %e\ncore-time: %f(secs)\n\n", 4.*count/NTRIAL, time);

    time = -omp_get_wtime();
    count = 0;
    for (size_t i = 0; i < NTRIAL; i++)
        x[i] = 2. * drand48() - 1.;
    for (size_t i = 0; i < NTRIAL; i++)
        y[i] = 2. * drand48() - 1.;
    for (size_t i = 0; i < NTRIAL; i++)
        z[i] = 1 - floor(x[i]*x[i] + y[i]*y[i]);
    for (size_t i = 0; i < NTRIAL; i++)
        count += z[i];
    time += omp_get_wtime();
    printf("value: %e\ncore-time: %f(secs)\n\n", 4.*count/NTRIAL, time);

    free(x);
    free(y);
    free(z);
    return 0;
}
